// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Prof. SONE's Laboratory, the Univ. of Tokyo.)
// PathSearchCell.cc --- class PathSearch functions
// $Id: PathSearchCell.cc,v 4.0.1.2 1996/07/08 08:38:56 rt Exp $

#include "PathSearchCell.hh"
#include <math.h>

// ֿ˹碌 tm_trans 򥻥åȤ
void
PathSearchCell::trf_set_trnum
( int trnum ) {
    register int i, j;
    tm_trans.setnum( trnum );
    for ( j = 0; j < trnum; j++ ) {
        tm_trans[ j ].setnum( trnum );
    }

    // ǥեȡå
    for ( i = 0; i < tm_trans.number(); i++ ) {
        for ( j = 0; j < tm_trans.number(); j++ ) {
            tm_trans[ i ][ j ] = 0;	// default: 
        }
    }
}    

// PathSearchCell 饹ξ费֥ơ֥ե뤫ɤ
void
PathSearchCell::rd_transfertime
( istream& str, string& fn ) {
    int iy, iz, iu;
    if ( getint ( str, iy ) == EOF ) {
        err_eof_noexit( fn ) << ": insufficient item(s)\n";
        exit( 1 );
    }
    if ( tm_trans.out_of_range( iy ) ) {
        err_eof_noexit( fn ) << ": train number " << iy
                             << " out of range\n";
        exit( 1 );
    }
    if ( getint ( str, iz ) == EOF ) {
        err_eof_noexit( fn ) << ": insufficient item(s)\n";
        exit( 1 );
    }
    if ( tm_trans[ iy ].out_of_range( iz ) ) {
        err_eof_noexit( fn ) << ": train number " << iz
                             << " out of range\n";
        exit( 1 );
    }
    if ( getinttime ( str, iu ) == EOF ) {
        err_eof_noexit( fn ) << ": insufficient item(s)\n";
        exit( 1 );
    }
    tm_trans[ iy ][ iz ] = iu;
    cerr << "Transfer time from train #" << iy << " to train #" << iz
         << ": " << iu << " sec(s). ";
}

// ѥץ: 
int sortsched_arr
( const void* ax, const void* bx ) {
    PathCell* a = ( PathCell* ) ax;
    PathCell* b = ( PathCell* ) bx;
    if ( a->tm_arr > b->tm_arr ) return 1;
    else if ( a->tm_arr < b->tm_arr ) return -1;
    else return 0;
}


// λɽ˥
void
StationTimeTable::sort_arrival
() {
    // ˤޤϰ˻ɽ
    for ( register int i = 0; i < number(); i++ ) {
        operator[]( i ).fit_in_cycle_arrival();
    }
    qsort( data, num, sizeof( PathCell ), &sortsched_arr );
}

// ѥץ
int sortsched_dep
( const void* ax, const void* bx ) {
    PathCell* a = ( PathCell* ) ax;
    PathCell* b = ( PathCell* ) bx;
    if ( a->tm_dep > b->tm_dep ) return 1;
    else if ( a->tm_dep < b->tm_dep ) return -1;
    else return 0;
}


// λɽ˥
void
StationTimeTable::sort_departure
() {
    // ˤޤϰ˻ɽ
    for ( register int i = 0; i < number(); i++ ) {
        operator[]( i ).fit_in_cycle_departure();
    }
    qsort( data, num, sizeof( PathCell ), &sortsched_dep );
}

int
StationTimeTable::find_by_id_tr
( int id_xtr ) {
    for ( register int i = 0; i < number(); i++ ) {
        if ( operator[]( i ).id_tr == id_xtr )
            return i;
    }
    cerr << "Warning: No StationTimeTable data found for train No. "
         << id_xtr << "\n";
    return -1;
}


//  i  j ؤξ费ˤ֡㤨פ֤Τȡ
int
TransferTable::operator()
( PathCell& cell_i, PathCell& cell_j ) {
    return operator()( cell_i.id_tr, cell_j.id_tr );
}


//  i  j ؤξ费ˤԤ
// 嵭Ρפ֡פ
// ͤǤʿ֤
int
TransferTable::waittime
( PathCell& cell_i, PathCell& cell_j ) {
    int tm_wait_tmp = cell_j.tm_dep - cell_i.tm_arr;
    tm_wait_tmp -= operator()( cell_i, cell_j );
    return tm_wait_tmp;
}


// 费ǽå
void
PathSearchCell::transferability_check
() {
    // ΰ
    transferable_forward.setnum( arrival.number() );
    register int i;
    for ( i = 0; i < arrival.number(); i++ ) {
        transferable_forward[ i ].renew();
    }
    transferable_backward.setnum( departure.number() );
    register int j;
    for ( j = 0; j < departure.number(); j++ ) {
        transferable_backward[ j ].renew();
    }

    PathCell cell_i;
    PathCell cell_j;	// i  j 
    int tm_wait_tmp = 0;

    // 
    for ( i = 0; i < arrival.number(); i++ ) {
        cell_i = arrival[ i ];		// ֤ξ费õ
        for ( j = 0; j < departure.number(); j++ ) {
            cell_j = departure[ j ];
            if ( cell_i == cell_j ) {
                // Ʊ: ٤碌о³Ϥ
                cell_j.adjust_cyc_tr( cell_i );
                transferable_forward[ i ] += cell_j;
                continue;	// for 롼פμ
            }
            do {
                tm_wait_tmp = tm_trans.waittime( cell_i, cell_j );
                if ( tm_wait_tmp <= 0 ) {
                    cell_j++;	// 1٤餻
                }
            } while ( tm_wait_tmp <= 0 );
            cell_j.tm_trf = tm_trans( cell_i, cell_j );
            cell_j.tm_wait = tm_wait_tmp;
            transferable_forward[ i ] += cell_j;
        }
    }

    // 
    for ( j = 0; j < departure.number(); j++ ) {
        cell_j = departure[ j ];		// ֤ؤξ费õ
        for ( i = 0; i < arrival.number(); i++ ) {
            cell_i = arrival[ i ];
            if ( cell_i == cell_j ) {
                // Ʊ: ٤碌о³Ϥ
                cell_i.adjust_cyc_tr( cell_j );
                transferable_backward[ j ] += cell_i;
                continue;	// for 롼פμ
            }
            do {
                tm_wait_tmp = tm_trans.waittime( cell_i, cell_j );
                if ( tm_wait_tmp <= 0 ) {
                    cell_i--;	// 1ʤޤ
                }
            } while ( tm_wait_tmp <= 0 );
            cell_i.tm_trf = tm_trans( cell_i, cell_j );
            cell_i.tm_wait = tm_wait_tmp;
            transferable_backward[ j ] += cell_i;
        }
    }
}


// ֥å
void
PathSearchCell::headway_check
() {
    // ΰ
    headway_departure.setnum( departure.number() );
    headway_arrival.setnum( arrival.number() );

    PathCell cell_i;
    PathCell cell_j;

    // ġνȯ֤Ȥλ֤Ǽ
    register int j;
    for ( j = 0; j < departure.number(); j++ ) {
        cell_j = departure[ j ];
        if ( j == 0 ) {
            cell_i = departure.last();
            cell_j++;
        } else {
            cell_i = departure[ j - 1 ];
        }
        headway_departure[ j ] = cell_j.tm_dep - cell_i.tm_dep;
    }

    // ġĸ֤Ȥλ֤Ǽ
    for ( j = 0; j < arrival.number(); j++ ) {
        cell_j = arrival[ j ];
        if ( j == arrival.number() - 1 ) {
            cell_i = arrival[ 0 ];
            cell_j--;
        } else {
            cell_i = arrival[ j + 1 ];
        }
        headway_arrival[ j ] = cell_i.tm_arr - cell_j.tm_arr;
    }
}


// arrival  PathCell ǡɲ
PathSearchCell&
PathSearchCell::operator+=
( PathCell& x ) {
    arrival += x;
    return *this;
}


// departure  PathCell ǡɲ
PathSearchCell&
PathSearchCell::operator*=
( PathCell& x ) {
    departure += x;
    return *this;
}


// arrival  departure ξ򥽡
void
PathSearchCell::sort
() {
    arrival.sort_arrival();
    departure.sort_departure();
}


// ػɽΰ
void
PathSearchCell::print_station_timetable
( ostream& str ) {
    string sepinit = ": ";
    string sepother = ", ";
    str << "   ***  ***\n";
    register int j;
    for ( j = 0; j < arrival.number(); j++ ) {
        str << "     " << setw(2) << setfill(' ') << arrival[ j ].id_tr << setw(0)
            << ":" << setw(5) << setfill(' ') << arrival[ j ].tm_arr << setw(0)
            << "(" << setw(2) << setfill(' ') << arrival[ j ].cyc_tr << setw(0)
            << ") [³";
        bool sw_sep = false;
        for ( register int k = 0;
              k < transferable_forward[ j ].number(); k++ ) {
            if ( transferable_forward[ j ][ k ].id_tr
                 != arrival[ j ].id_tr ) {
                if ( sw_sep ) {
                    str << sepother;
                } else {
                    str << sepinit;
                    sw_sep = true;
                }
                str << transferable_forward[ j ][ k ].id_tr << "("
                    << transferable_forward[ j ][ k ].tm_dep << "/"
                    << transferable_forward[ j ][ k ].cyc_tr << ")";
            }
        }
        if ( ! sw_sep ) {
            str << "ʤ";
        }
        str << "]\n";
    }
    if ( arrival.number() == 0 ) {
        str << "        \n";
    }
    str << "   *** ȯ ***\n";
    for ( j = 0; j < departure.number(); j++ ) {
        str << "     " << setw(2) << setfill(' ') << departure[ j ].id_tr << setw(0)
            << ":" << setw(5) << setfill(' ') << departure[ j ].tm_dep << setw(0)
            << "(" << setw(2) << setfill(' ') << departure[ j ].cyc_tr << setw(0)
            << ") [³";
        bool sw_sep = false;
        for ( register int k = 0;
              k < transferable_backward[ j ].number(); k++ ) {
            if ( transferable_backward[ j ][ k ].id_tr
                 != departure[ j ].id_tr ) {
                if ( sw_sep ) {
                    str << sepother;
                } else {
                    str << sepinit;
                    sw_sep = true;
                }
                str << transferable_backward[ j ][ k ].id_tr << "("
                    << transferable_backward[ j ][ k ].tm_arr << "/"
                    << transferable_backward[ j ][ k ].cyc_tr << ")";
            }
        }
        if ( ! sw_sep ) {
            str << "ʤ";
        }
        str << "]\n";
    }
    if ( departure.number() == 0 ) {
        str << "        \n";
    }
}


// ˤ费ǽο
//  pcl Ф³ȯ֥ǡ
int
PathSearchCell::forward_connections
( PathCell& pcl ) {
    return transferable_forward[ arrival( pcl ) ].number();
}


// ˤ费ǽ
//  pcl Ф³ ix ܤνȯ
// arrival äƤ PathCell ǡ
// transferable_forward ơ֥äƤ PathCell ǡˤƤ
// ٤³Ȥ褦ˤƤΤǡ
// ֤ǡ⤳˹碌
PathCell
PathSearchCell::forward_connection
( PathCell pcl, int ix ) {

    // pcl Ф ix ³ȯ֤
    PathCell ret_pc = transferable_forward[ arrival( pcl ) ][ ix ];

    // pcl  arrival ǡȤμ٤ܿκ
    int cyc_inc = arrival.cyc_difference( pcl );

    // ٤ܿκ ret_pc ˲ä
    ret_pc += cyc_inc;

    return ret_pc;
}


// ˤ费ǽο
//  pcl Ф³֥ǡ
int
PathSearchCell::backward_connections
( PathCell& pcl ) {
    return transferable_backward[ departure( pcl ) ].number();
}


// ˤ费ǽ
// ȯ pcl Ф³ ix ܤ
// departure äƤ PathCell ǡ
// transferable_backward ơ֥äƤ PathCell ǡˤƤ
// ٤³Ȥ褦ˤƤΤǡ
// ֤ǡ⤳˹碌
PathCell
PathSearchCell::backward_connection
( PathCell pcl, int ix ) {

    // pcl Ф ix ³֤
    PathCell ret_pc = transferable_backward[ departure( pcl ) ][ ix ];

    // pcl  departure ǡȤμ٤ܿκ
    int cyc_inc = departure.cyc_difference( pcl );

    // ٤ܿκ ret_pc ˲ä
    ret_pc += cyc_inc;

    return ret_pc;
}


// ȯ֤֤Υѥ
PathCell
PathSearchCell::departingtrain
( int i ) {
    return departure[ i ];
}


// 夹֤Υѥ
PathCell
PathSearchCell::arrivingtrain
( int i ) {
    return arrival[ i ];
}


// 费ҿΥå
void
PathSearchCell::arrival_trf_num
( const PathCell& xpc, int ix )
{
    arrival[ arrival_id( xpc ) ].num_passen += ix;
}


// 费ҿΥå
void
PathSearchCell::departure_trf_num
( const PathCell& xpc, int ix )
{
    departure[ departure_id( xpc ) ].num_passen += ix;
}


// 费ҿŪ˼Ф
const StationTimeTable&
PathSearchCell::get_transfer_numbers
() {
    static StationTimeTable ret_val;
    ret_val.renew();
    ret_val = arrival;
    for ( int j = 0; j < departure.number(); j++ ) {
        PathCell xp = departure[ j ];
        xp.num_passen *= -1;
        xp.tm_arr = xp.tm_dep;
        ret_val += xp;
    }
    ret_val.sort_arrival();
    return ret_val;
}
