// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Prof. SONE's Laboratory, the Univ. of Tokyo.)
// PathSearchCell.hh --- header file that defines class PathSearchCell
// $Id: PathSearchCell.hh,v 4.0.1.2 1996/07/08 08:38:56 rt Exp $

#include "PathMap.hh"
#include "RefTable.hh"
#include <iomanip.h>

#ifndef ___PATHSEARCHCELL_HH
#define ___PATHSEARCHCELL_HH


class StationTimeTable: public PathMap {
  public:
    void sort_arrival();	// Ȥ¦ѡ
    void sort_departure();	// Ȥʽȯ¦ѡ
    int find_by_id_tr( int );	/* id_tr  StationTimeTable 
                                   ܤ򥵡֤ */
};

//  i  j ؤξ费ˤ
class TransferTableY: public RefTable< int > {
  private:
    void error26_name() const {
        cerr << "Error: subscript out of range (TransferTableY)\n";
    }
};


class TransferTable: public RefTable< RefTable< int > > {
  public:
    int operator()( PathCell&, PathCell& );	// 费֤֤
    int operator()( int i, int j )		// 费֤֤
        {return operator[]( i )[ j ];}
    int waittime( PathCell&, PathCell& );	// Ԥ֤֤

  private:
    void error26_name() const {
        cerr << "Error: subscript out of range (TransferTableX)\n";
    }
};


class TransferAbilityY: public RefTable< PathCell > {
  private:
    void error26_name() const {
        cerr << "Error: subscript out of range (TransferAbilityY)\n";
    }
};
    

class TransferAbility: public RefTable< TransferAbilityY > {
  private:
    void error26_name() const {
        cerr << "Error: subscript out of range (TransferAbilityX)\n";
    }
};
    

class HeadwayTable: public RefTable< int > {
  private:
    void error26_name() const {
        cerr << "Error: subscript out of range (HeadwayTable)\n";
    }
};
    

// ѥΤΥǡ
class PathSearchCell {

  protected:
    StationTimeTable arrival;		// ɽ
    StationTimeTable departure;		// ȯֻɽ

    TransferAbility			// ˤ
    transferable_forward;		// 费ǽơ֥

    TransferAbility			// ˤ
    transferable_backward;		// 费ǽơ֥

    TransferTable tm_trans;		// 费ˤ֥ơ֥

    HeadwayTable headway_departure;	// ȯߤλ
    HeadwayTable headway_arrival;	// ߤλ

  public:
    PathSearchCell()
    : arrival(), departure(), tm_trans() {}

    PathSearchCell( const PathSearchCell& x )
    : arrival( x.arrival ), departure( x.departure ),
      tm_trans( x.tm_trans ) {}

    PathSearchCell& operator=( const PathSearchCell& x ) {
        if ( &x == this ) return *this;
        membercopy( x );
        return *this;
    }

    // Ǿ费֥ơ֥ tm_trans طΥ˥饤
    void trf_set_trnum( int );		// 󥻥å
    void path_set_pathtblnum_forward();	// 󥻥å
    void path_set_pathtblnum_backward();	// 󥻥å
    void rd_transfertime( istream&, string& );	// ե뤫ɤ߹

    // ػɽ
    void sort();
    void headway_check();		// ֥å
    void transferability_check();	// 费ǽΥå

    // ػɽФ
    PathCell departingtrain( int i );	// ȯ֤֤Υѥ
    PathCell arrivingtrain( int i );	// 夹֤Υѥ
    int departures()
        {return departure.number();}	// ȯֿ֤
    int arrivals()
        {return arrival.number();}	// 夹ֿ
    int departure_id( const PathCell& pc )	// ȯ֤ pc 
        {return departure( pc );}		// ȯֹ
    int arrival_id( const PathCell& pc )	// 夹 pc 
        {return arrival( pc );}			// ֹ
    const HeadwayTable& departure_headways()
        {return headway_departure;}	// ȯߴ֤λ
    const HeadwayTable& arrival_headways()
        {return headway_arrival;}	// ߴ֤λ

    // ˤ费ǽ֤³ȯ֡ˤθ
    int forward_connections		// Ϳ֤³
    ( PathCell& );			// ȯֿ
    PathCell forward_connection		// Ϳ֤³
    ( PathCell, int );			//  (int) ܤνȯ

    // ˤ费ǽ
    int backward_connections		// Ϳ֤³
    ( PathCell& );			// ֿ
    PathCell backward_connection	// Ϳ֤³
    ( PathCell, int );			//  (int) ܤ

    // 费ҿΥå
    void arrival_trf_num( const PathCell&, int );
    void departure_trf_num( const PathCell&, int );

    // 费ҿŪ˼Ф
    const StationTimeTable& get_transfer_numbers();

  private:
    void membercopy( const PathSearchCell& x ) {
        arrival = x.arrival;	departure = x.departure;
        tm_trans = x.tm_trans;
        transferable_forward = x.transferable_forward;
        transferable_backward = x.transferable_backward;
    }        

  public:
    PathSearchCell& operator+=		// arrival  PathCell ѿ
    ( PathCell& );			// ɲ
    PathSearchCell& operator*=		// departure  PathCell ѿ
    ( PathCell& );			// ɲ

    void print_station_timetable( ostream& str );

    friend istream& operator>> ( istream&, PathSearchCell& );
    friend ostream& operator<< ( ostream&, PathSearchCell& );
};

#endif /* ! ___PATHSEARCHCELL_HH */
